﻿using System.Linq;
using ComponentFramework;
using ComponentFramework.Modules;
using IrrKlang;
using MTV3D65;
using System;
using TroubleInEuclidea.Structure;
using TroubleInEuclidea.Services;
using TroubleInEuclidea.Components.Enemies;

namespace TroubleInEuclidea.Components
{
    [AutoLoad]
    class Player : Component, IShip
    {
        const float MovementSpeed = 1.5f;
        const float Friction = 0.8f;

        TVMesh mesh;
        TimeSpan sinceDied, sinceSpawned;
        bool flash;
        ISound boosterSound;
        Enemy hitEnemy;

        public Player(ICore core) : base(core) { }

        public override void Initialize()
        {
            mesh = SpriteFactory.CreateSprite("Player");
            mesh.SetScale(Size.x, Size.y, 1);

            boosterSound = SoundManager.GetSound(SoundManager.Load("booster"), true, false);
            sinceDied = TimeSpan.FromSeconds(3);
        }

        public override void Update(TimeSpan elapsedTime)
        {
            if (GameState.Paused) return;

            if (PlayerState.Alive)
            {
                var movement = new TV_2DVECTOR();

                if (InputEngine.IsKeyPressed(CONST_TV_KEY.TV_KEY_W))
                    movement.y++;
                if (InputEngine.IsKeyPressed(CONST_TV_KEY.TV_KEY_S))
                    movement.y--;
                if (InputEngine.IsKeyPressed(CONST_TV_KEY.TV_KEY_D))
                    movement.x++;
                if (InputEngine.IsKeyPressed(CONST_TV_KEY.TV_KEY_A))
                    movement.x--;

                boosterSound.Paused = movement.AlmostEqual(Vector2.Zero);

                movement = movement.Normalize();
                movement *= MovementSpeed * (float)elapsedTime.TotalSeconds;

                PlayerState.Velocity += movement;
                PlayerState.Velocity *= Friction;

                PlayerState.Position += PlayerState.Velocity;
                PlayerState.Position = PlayerState.Position.Clamp(LevelState.Size / 2 - Vector2.One / 2);

                int mx = 0, my = 0;
                InputEngine.GetMouseState(ref mx, ref my);
                PlayerState.Heading = PlayerState.Heading.Lerp(PlayerState.Heading + new TV_2DVECTOR(mx, -my) * 1.75f, 0.01).Normalize();

                // Hit a ship?
                if (PlayerState.Ready && (hitEnemy = CollisionManager.OverlapsAny(PlayerState.Position, Vector2.One)) != null) 
                {
                    PlayerState.Alive = false;
                    PlayerState.HasNotDied = false;
                    PlayerState.Lives--;

                    SoundManager.Play(SoundManager.Load("boom1"), false);
                    if (PlayerState.Lives > 0)
                        SoundManager.Play(SoundManager.Load("buzz8"), false);
                    else
                        SoundManager.Play(SoundManager.Load("buzz6"), false);

                    if (PlayerState.Lives != 0)
                        GameState.DowngradeDifficulty();

                    foreach (var enemy in EnemiesState.Enemies.Where(x => x != hitEnemy))
                        Core.UnloadComponent(enemy);
                    EnemiesState.Enemies.Clear();
                    hitEnemy.Velocity = Vector2.Zero;

                    sinceDied = TimeSpan.Zero;
                    Core.LoadComponent(new ExplodingShip(Core, this));
                }

                mesh.SetPosition(PlayerState.Position.x, PlayerState.Position.y, 0);
                mesh.SetRotation(0, 0, PlayerState.Theta + MathHelper.HalfPi);
            }
            else
            {
                boosterSound.Paused = true;

                sinceDied += elapsedTime;
                if (sinceDied.TotalSeconds > 3)
                {
                    if (PlayerState.Lives > 0 || InputEngine.IsKeyPressed(CONST_TV_KEY.TV_KEY_SPACE))
                    {
                        if (PlayerState.Lives == 0)
                        {
                            PlayerState.Lives = 4;
                            PlayerState.Score = 0;
                            PlayerState.HasNotDied = true;
                            GameState.Difficulty = DifficultyLevel.Noob;
                            GameState.DifficultyTimer = TimeSpan.Zero;
                        }

                        PlayerState.Velocity = Vector2.Zero;
                        sinceSpawned = TimeSpan.Zero;
                        PlayerState.Alive = true;
                        PlayerState.Ready = false;
                        if (hitEnemy != null)
                            Core.UnloadComponent(hitEnemy);

                        SoundManager.Play(SoundManager.Load("beep4"));
                    } 
                        
                }
            }

            if (!PlayerState.Ready)
            {
                flash = Math.Round(sinceSpawned.TotalMilliseconds / 150) % 2 == 0;
                sinceSpawned += elapsedTime;

                if (sinceSpawned.TotalSeconds > 1)
                    PlayerState.Ready = true;
            }

#if DEBUG
            if (InputEngine.IsKeyPressed(CONST_TV_KEY.TV_KEY_1))
                GameState.Difficulty = DifficultyLevel.Noob;
            if (InputEngine.IsKeyPressed(CONST_TV_KEY.TV_KEY_2))
                GameState.Difficulty = DifficultyLevel.Amateur;
            if (InputEngine.IsKeyPressed(CONST_TV_KEY.TV_KEY_3))
                GameState.Difficulty = DifficultyLevel.Skilled;
            if (InputEngine.IsKeyPressed(CONST_TV_KEY.TV_KEY_4))
                GameState.Difficulty = DifficultyLevel.Addict;
            if (InputEngine.IsKeyPressed(CONST_TV_KEY.TV_KEY_5))
                GameState.Difficulty = DifficultyLevel.Gentleman;
            if (InputEngine.IsKeyPressed(CONST_TV_KEY.TV_KEY_6))
                GameState.Difficulty = DifficultyLevel.Korean;

            if (InputEngine.IsKeyPressed(CONST_TV_KEY.TV_KEY_F1))
                PlayerState.Lives = 1;
#endif
        }

        public override void Draw()
        {
            if (PlayerState.Alive && (PlayerState.Ready || flash))
            {
                Scene.SetTextureFilter(CONST_TV_TEXTUREFILTER.TV_FILTER_NEARESTPOINT);
                mesh.Render();
            }
        }

        public TV_2DVECTOR Position
        {
            get { return PlayerState.Position; }
        }
        public TV_2DVECTOR Velocity
        {
            get { return PlayerState.Velocity; }
        }
        public TV_2DVECTOR Size
        {
            get { return new TV_2DVECTOR(0.9f, 0.9f); }
        }
        public TV_COLOR Color
        {
            get { return new TV_COLOR(1, 1, 1, 1); }
        }

        [ServiceDependency]
        public IPlayerState PlayerState { private get; set; }
        [ServiceDependency]
        public IEnemiesState EnemiesState { private get; set; }
        [ServiceDependency]
        public IGameState GameState { private get; set; }
        [ServiceDependency]
        public ICollisionManager CollisionManager { private get; set; }
        [ServiceDependency]
        public ILevelState LevelState { private get; set; }
        [ServiceDependency]
        public ISoundManager SoundManager { private get; set; }
    }
}
