﻿using ComponentFramework;
using ComponentFramework.Modules;
using MTV3D65;
using TroubleInEuclidea.Services;
using TroubleInEuclidea.Structure;
using System;
using System.Collections.Generic;
using TroubleInEuclidea.Components.Enemies;

namespace TroubleInEuclidea.Components
{
    [AutoLoad]
    class Weapon : Component
    {
        const int InstanceCount = 104;
        const float BulletSpeed = 10;
        const float BulletFrequency = 0.075f;
        static readonly TV_2DVECTOR BulletSize = new TV_2DVECTOR(0.4f, 0.5f);

        TimeSpan timeSinceLastShot;
        TVMiniMesh instances;
        TVMesh template;

        readonly List<BulletState> bullets = new List<BulletState>();
        int nextBulletId;

        public Weapon(ICore core) : base(core) {}

        public override void Initialize()
        {
            template = SpriteFactory.CreateSprite("Bullet");
            template.SetScale(BulletSize.x, BulletSize.y, 1);

            instances = Scene.CreateMiniMesh(InstanceCount);
            instances.CreateFromMesh(template);
            instances.SetColor(Globals.RGBA(1, 0.992f, 0.49f, 1));
            instances.SetColorMode(true);
        }

        public override void Update(TimeSpan elapsedTime)
        {
            if (GameState.Paused) return;

            timeSinceLastShot += elapsedTime;

            // Check for new shots
            if (PlayerState.Alive && PlayerState.Ready && InputEngine.IsMouseButtonPressed(0) && timeSinceLastShot.TotalSeconds > BulletFrequency)
            {
                bullets.Add(new BulletState { Id = nextBulletId, Direction = PlayerState.Heading, InitialVelocity = PlayerState.Velocity * 0.5f });

                instances.EnableMiniMesh(nextBulletId, true);
                TV_2DVECTOR position = PlayerState.Position + PlayerState.Heading / 2;
                instances.SetPosition(position.x, position.y, 0, nextBulletId);
                instances.SetRotation(0, 0, PlayerState.Theta, nextBulletId);

                nextBulletId++;
                if (nextBulletId == InstanceCount)
                    nextBulletId = 0;

                timeSinceLastShot = TimeSpan.Zero;

                SoundManager.Play(SoundManager.Load("buzz10"));
            }

            // Update existing shots
            foreach (BulletState bullet in bullets.ToArray())
            {
                TV_2DVECTOR position = instances.GetPosition(bullet.Id).ToVector2();
                TV_2DVECTOR velocity = bullet.InitialVelocity + bullet.Direction * BulletSpeed * (float)elapsedTime.TotalSeconds;
                position += velocity;

                // Went out of bounds?
                if (!position.IsInside(LevelState.Size / 1.8f))
                {
                    instances.EnableMiniMesh(bullet.Id, false);
                    bullets.Remove(bullet);
                    return;
                }

                // Hit an enemy?
                Enemy collided = CollisionManager.OverlapsAny(position, BulletSize);
                if (collided != null)
                {
                    Core.UnloadComponent(collided);
                    EnemiesState.Enemies.Remove(collided);

                    PlayerState.Score += collided.Points;
                    Core.LoadComponent(new ExplodingShip(Core, collided));
                }

                instances.SetPosition(position.x, position.y, 0, bullet.Id);
            }
        }

        public override void Draw()
        {
            instances.Render();
        }

        class BulletState
        {
            public int Id { get; set; }
            public TV_2DVECTOR Direction { get; set; }
            public TV_2DVECTOR InitialVelocity { get; set; }
        }

        [ServiceDependency]
        public IGameState GameState { private get; set; }
        [ServiceDependency]
        public IPlayerState PlayerState { private get; set; }
        [ServiceDependency]
        public IEnemiesState EnemiesState { private get; set; }
        [ServiceDependency]
        public ILevelState LevelState { private get; set; }
        [ServiceDependency]
        public ICollisionManager CollisionManager { private get; set; }
        [ServiceDependency]
        public ISoundManager SoundManager { private get; set; }
    }
}
